/*
*
*    OPEN-XCHANGE legal information
*
*    All intellectual property rights in the Software are protected by
*    international copyright laws.
*
*
*    In some countries OX, OX Open-Xchange, open xchange and OXtender
*    as well as the corresponding Logos OX Open-Xchange and OX are registered
*    trademarks.
*    The use of the Logos is not covered by the GNU General Public License.
*    Instead, you are allowed to use these Logos according to the terms and
*    conditions of the Creative Commons License, Version 2.5, Attribution,
*    Non-commercial, ShareAlike, and the interpretation of the term
*    Non-commercial applicable to the aforementioned license is published
*    on the web site http://www.open-xchange.com/EN/legal/index.html.
*
*    Please make sure that third-party modules and libraries are used
*    according to their respective licenses.
*
*    Any modifications to this package must retain all copyright notices
*    of the original copyright holder(s) for the original code used.
*
*    After any such modifications, the original and derivative code shall remain
*    under the copyright of the copyright holder(s) and/or original author(s)per
*    the Attribution and Assignment Agreement that can be located at
*    http://www.open-xchange.com/EN/developer/. The contributing author shall be
*    given Attribution for the derivative code and a license granting use.
*
*     Copyright (C) 2016 OX Software GmbH
*     Mail: info@open-xchange.com
*
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License, Version 2 as published
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc., 59
*     Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*/
package com.openexchange.office.hazelcast.osgi;

import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.osgi.framework.BundleException;

import com.hazelcast.config.Config;
import com.hazelcast.config.MapConfig;
import com.hazelcast.core.HazelcastInstance;
import com.openexchange.exception.OXException;
import com.openexchange.filemanagement.ManagedFileManagement;
import com.openexchange.management.ManagementService;
import com.openexchange.office.hazelcast.access.HazelcastAccess;
import com.openexchange.office.hazelcast.cleanup.CleanupMembershipListener;
import com.openexchange.office.hazelcast.doc.DocSaveStateDirectory;
import com.openexchange.office.hazelcast.doc.DocumentDirectory;
import com.openexchange.office.hazelcast.doc.DocumentResourcesDirectory;
import com.openexchange.office.hazelcast.doc.HazelcastDocSaveStateDirectory;
import com.openexchange.office.hazelcast.doc.HazelcastDocumentCleanupLock;
import com.openexchange.office.hazelcast.doc.HazelcastDocumentDirectory;
import com.openexchange.office.hazelcast.doc.HazelcastDocumentResourcesDirectory;
import com.openexchange.office.hazelcast.management.ManagementHouseKeeper;
import com.openexchange.osgi.HousekeepingActivator;
import com.openexchange.threadpool.ThreadPoolService;
import com.openexchange.timer.TimerService;

public class HazelcastOfficeActivator extends HousekeepingActivator {

    @SuppressWarnings("deprecation")
    private final Log LOG = com.openexchange.log.Log.loggerFor(HazelcastOfficeActivator.class);

    private final AtomicBoolean isStopped = new AtomicBoolean(true);
    private volatile HazelcastDocumentDirectory docStateDir;
    private volatile HazelcastDocSaveStateDirectory docSaveStateDir;
    private volatile HazelcastDocumentResourcesDirectory docResourcesDir;

    @Override
    protected Class<?>[] getNeededServices() {
        return new Class<?>[]
            {
                HazelcastInstance.class,
                ManagementService.class,
                TimerService.class,
                ThreadPoolService.class,
                ManagedFileManagement.class
            };
    }

	@Override
    protected void startBundle() throws Exception {
       LOG.debug("Starting bundle: " + getClass().getCanonicalName());
       Services.setServiceLookup(this);

       final ManagementHouseKeeper managementHouseKeeper = ManagementHouseKeeper.getInstance();
       managementHouseKeeper.initialize(this);

       final HazelcastInstance hazelcastInstance = getService(HazelcastInstance.class);
       HazelcastAccess.setHazelcastInstance(hazelcastInstance);
       final Config config = hazelcastInstance.getConfig();

       final String docStateResource_map = discoverMapName(config, "officeDocumentDirectory-");
       if(StringUtils.isEmpty(docStateResource_map)) {
           final String msg = "Distributed documents persistent state directory map couldn't be found in hazelcast configuration";
           throw new IllegalStateException(msg, new BundleException(msg, BundleException.ACTIVATOR_ERROR));
       }
       LOG.debug("Registered officeDocumentDirectory to Hazelcast");

       final String docSaveStateResource_map = discoverMapName(config, "officeDocSaveStateDirectory-");
       if(StringUtils.isEmpty(docSaveStateResource_map)) {
           final String msg = "Distributed document save state directory map couldn't be found in hazelcast configuration";
           throw new IllegalStateException(msg, new BundleException(msg, BundleException.ACTIVATOR_ERROR));
       }
       LOG.debug("Registered officeDocSaveStateDirectory to Hazelcast");

       final String docResources_map = discoverMapName(config, "officeDocumentResources-");
       if(StringUtils.isEmpty(docResources_map)) {
           final String msg = "Distributed document resources directory map couldn't be found in hazelcast configuration";
           throw new IllegalStateException(msg, new BundleException(msg, BundleException.ACTIVATOR_ERROR));
       }
       LOG.debug("Registered officeDocumentResources to Hazelcast");

       String lock_map = discoverMapName(config, "officeCleanupLock-");

       // register our hazelcast structures
       try {
    	   docStateDir = new HazelcastDocumentDirectory(docStateResource_map);
    	   docSaveStateDir = new HazelcastDocSaveStateDirectory(docSaveStateResource_map);
    	   docResourcesDir = new HazelcastDocumentResourcesDirectory(docResources_map);
    	   HazelcastDocumentCleanupLock lockDir = new HazelcastDocumentCleanupLock(lock_map);

           managementHouseKeeper.addManagementObject(docSaveStateDir.getManagementObject());
           registerService(DocSaveStateDirectory.class, docSaveStateDir, null);

           managementHouseKeeper.addManagementObject(docStateDir.getManagementObject());
           registerService(DocumentDirectory.class, docStateDir, null);

           managementHouseKeeper.addManagementObject(docResourcesDir.getManagementObject());
           registerService(DocumentResourcesDirectory.class, docResourcesDir, null);

           managementHouseKeeper.addManagementObject(lockDir.getManagementObject());
       } catch (Throwable e) {
           LOG.error("Failed to create/register hazelcast document directories", e);
       }

       // add cleanup listener to remove entries controlled by a cluster member that is shutdown
       final CleanupMembershipListener cleanupListener = new CleanupMembershipListener(lock_map, docStateDir, docResourcesDir);
       hazelcastInstance.getCluster().addMembershipListener(cleanupListener);

       try {
           managementHouseKeeper.exposeManagementObjects();
       } catch (OXException oxe) {
           LOG.error("Failed to expose ManagementObjects", oxe);
       }

       openTrackers();
       isStopped.set(false);
	}

	@Override
    public void stopBundle() throws Exception {
        if (isStopped.compareAndSet(false, true)) {
            LOG.debug("Stopping bundle: " + getClass().getCanonicalName());
            ManagementHouseKeeper.getInstance().cleanup();

            Services.setServiceLookup(null);
            cleanUp();
            super.stopBundle();
        }
	}

   @Override
   protected void handleAvailability(Class<?> clazz) {
       if (allAvailable()) {
           LOG.debug("HazelcastOfficeActivator regained needed service " + clazz.getSimpleName() + ". Going to start bundle.");
           try {
               startBundle();
           } catch (Exception e) {
               LOG.error("Error while starting bundle.", e);
           }
       }
   }

   @Override
   protected void handleUnavailability(Class<?> clazz) {
       LOG.warn("HazelcastOfficeActivator is handling unavailibility of needed service " + clazz.getSimpleName() + ". Going to stop bundle.");
       try {
           this.stopBundle();
       } catch (Exception e) {
           LOG.error("Error while stopping bundle.", e);
       }
   }

   /**
    * Discovers map names in the supplied hazelcast configuration based on the map prefix.
    *
    * @param config The config object
    * @return The prefix of the map name
    */
   private String discoverMapName(Config config, String mapPrefix){
       Map<String, MapConfig> mapConfigs = config.getMapConfigs();
       if (null != mapConfigs && 0 < mapConfigs.size()) {
           for (String mapName : mapConfigs.keySet()) {
               if (mapName.startsWith(mapPrefix)) {
                   return mapName;
               }
           }
       }
       return null;
   }
}
